
import logging

import click


from .copy import cli_copy
from .manage import cli_manage
from .download import cli_download
from .upload import cli_upload, cli_upload_advanced
from .user import cli_user
from .view import cli_view
from .search import cli_search
from geoseeq.vc.cli import cli_vc
from geoseeq.knex import DEFAULT_ENDPOINT
from geoseeq.utils import set_profile
from .shared_params.opts_and_args import overwrite_option
from .detail import cli_detail
from .run import cli_app
from .get_eula import cli_eula

logger = logging.getLogger('geoseeq_api')
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('[%(levelname)s] %(name)s :: %(message)s'))
logger.addHandler(handler)


@click.group()
def main():
    """Command line interface for the GeoSeeq API.
    
    ---

    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    pass

main.add_command(cli_download)
main.add_command(cli_upload)
main.add_command(cli_manage)
main.add_command(cli_view)
main.add_command(cli_search)
main.add_command(cli_app)
main.add_command(cli_eula)

@main.command()
def version():
    """Print the version of the Geoseeq API being used.

    ---
    
    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    click.echo('0.5.1')  # remember to update setup


@main.group('advanced')
def cli_advanced():
    """Advanced commands."""
    pass

cli_advanced.add_command(cli_copy)
cli_advanced.add_command(cli_user)
cli_advanced.add_command(cli_detail)
cli_advanced.add_command(cli_upload_advanced)

@cli_advanced.group('experimental')
def cli_experimental():
    """Experimental commands."""
    pass

cli_experimental.add_command(cli_vc)

@main.command('config')
@click.option('-p', '--profile', default=None, help='The profile name to use.')
@overwrite_option
def cli_config(profile, overwrite):
    """Configure the GeoSeeq API.

    ---
    
    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    if not profile:
        profile = click.prompt(f'Set custom profile name? (Leave blank for default)', default="").strip(' \"\'')
    endpoint = click.prompt(f'Enter the URL to use for GeoSeeq (Most users can use the default)', default=DEFAULT_ENDPOINT).strip(' \"\'')
    api_token = click.prompt(f'Enter your GeoSeeq API token', hide_input=True).strip(' \"\'')
    eula_accepted = click.confirm(f'Have you read and accepted the GeoSeeq End User License Agreement? Use `geoseeq eula show` to view the EULA.')
    if not eula_accepted:
        click.echo('You must accept the EULA to use the GeoSeeq API.')
        return
    set_profile(api_token, endpoint=endpoint, profile=profile, overwrite=overwrite)
    click.echo(f'Profile configured.')


import hashlib
import os
import logging
from ftplib import FTP
from threading import Timer
from .file_system_cache import FileSystemCache
from os.path import join, exists
import json
from os import environ, makedirs
from .constants import CONFIG_DIR, PROFILES_PATH, DEFAULT_ENDPOINT

logger = logging.getLogger('geoseeq_api')  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def load_auth_profile(profile=""):
    """Return an endpoit and a token"""
    profile = profile or "__default__"
    with open(PROFILES_PATH, "r") as f:
        profiles = json.load(f)
    if profile in profiles:
        return profiles[profile]["endpoint"], profiles[profile]["token"]
    raise KeyError(f"Profile {profile} not found.")


def set_profile(token, endpoint=DEFAULT_ENDPOINT, profile="", overwrite=False):
    """Write a profile to a config file.
    
    Raises KeyError if profile already exists.
    """
    if not exists(PROFILES_PATH):
        makedirs(CONFIG_DIR)
        with open(PROFILES_PATH, "w") as f:
            json.dump({}, f)
    with open(PROFILES_PATH, "r") as f:
        profiles = json.load(f)
    profile = profile or "__default__"
    if profile in profiles and not overwrite:
        raise KeyError(f"Profile {profile} already exists.")
    profiles[profile] = {
        "token": token,
        "endpoint": endpoint,
    }
    with open(PROFILES_PATH, "w") as f:
        json.dump(profiles, f, indent=4)


def paginated_iterator(knex, initial_url, error_handler=None):
    cache = FileSystemCache()
    result = cache.get_cached_blob(initial_url)
    if not result:
        try:
            result = knex.get(initial_url)
        except Exception as e:
            logger.debug(f'Error fetching blob:\n\t{initial_url}\n\t{e}')
            if error_handler:
                error_handler(e)
            else:
                raise
        cache.cache_blob(initial_url, result)
    for blob in result['results']:
        yield blob
    next_page = result.get('next', None)
    if next_page:
        for blob in paginated_iterator(knex, next_page):
            yield blob


def md5_checksum(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



def download_ftp(url, local_file_path, head=None):
    tkns = url.split('ftp://')[1].split('/')
    host, folder_path, file_name = tkns[0], '/'.join(tkns[1:-1]), tkns[-1]
    ftp = FTP(host)
    ftp.login()
    ftp.cwd(folder_path)
    logger.debug(f'Logged into {host} and changed to {folder_path}')

    def handle_download(file_handle, block):
        logger.debug(f'Writing block of size {len(block)} to {local_file_path}')
        file_handle.write(block)
        if head and os.path.getsize(local_file_path) >= head:
            logger.debug(f'File {local_file_path} has reached head size of {head} bytes. Aborting download.')
            ftp.sock.close()
            with open(local_file_path, 'rb+') as f:
                f.seek(head)
                f.truncate()
            assert False  # this is the best way to abort a download
        
    with open(local_file_path, 'wb') as f:
        try:
            ftp.retrbinary('RETR ' + file_name, lambda block: handle_download(f, block))
        except AssertionError:
            pass
    ftp.close()

    # trim local file to head size
    if head:
        with open(local_file_path, 'rb+') as f:
            f.seek(head)
            f.truncate()



import click
from geoseeq.knex import DEFAULT_ENDPOINT

from .. import Knex, Organization


@click.group('copy')
def cli_copy():
    """Copy samples and analysis results from one geoseeq instance to another.
    
    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    pass


@cli_copy.command('group')
@click.option('-l', '--log-level', type=int, default=20, envvar='GEOSEEQ_CLI_LOG_LEVEL')
@click.option('-o', '--outfile', type=click.File('w'), default='-')
@click.option('--source-api-token', envvar='SOURCE_GEOSEEQ_API_TOKEN')
@click.option('--target-api-token', envvar='GEOSEEQ_API_TOKEN')
@click.option('--source-endpoint', default=DEFAULT_ENDPOINT)
@click.option('--target-endpoint', default="https://geoseeq.dev.biotia.io")
@click.argument('source_org_name')
@click.argument('source_grp_name')
@click.argument('target_org_name')
@click.argument('target_grp_name')
def cli_list_samples(log_level, outfile,
                     source_api_token, target_api_token,
                     source_endpoint, target_endpoint,
                     source_org_name, source_grp_name,
                     target_org_name, target_grp_name):
    """Copy a group and its samples from one geoseeq instance to another.
    
    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    source_knex = Knex(source_endpoint)
    if source_api_token:
        source_knex.add_api_token(source_api_token)
    source_org = Organization(source_knex, source_org_name).get()
    source_grp = source_org.sample_group(source_grp_name).get()

    target_knex = Knex(target_endpoint)
    if target_api_token:
        target_knex.add_api_token(target_api_token)
    target_org = Organization(target_knex, target_org_name).idem()
    target_grp = target_org.sample_group(target_grp_name).idem()

    for source_ar in source_grp.get_analysis_results():
        target_ar = source_ar.copy(target_grp, save=True)
        print(source_ar, target_ar, file=outfile)

    for source_sample in source_grp.get_samples():
        target_sample = source_sample.copy(target_grp, save=True)
        print(source_sample, target_sample, file=outfile)



from geoseeq.result import ResultFile
from typing import Literal


class BaseLayer:

    def __init__(self,
                 title,
                 type: Literal["vector-tile", "raster-tile", "image"],
                 source=None,
                 access_token=None,
                 info=""):
        self.title = title
        self.type = type
        self.source = source
        self.access_token = access_token
        self.info = info

    def set_source(self, source, access_token=None):
        self.source = source
        if access_token:
            self.access_token = access_token
        return self

    def set_info(self, info):
        self.info = info
        return self
    
    def set_bounds(self, lat_n, lon_w, lat_s, lon_e):
        self.bounds = [
            [lat_n, lon_w],
            [lat_s, lon_e]
        ]
        return self
    
    def set_legend(self, label, value_range):
        self.legend = {
            "label": label,
            "valueRange": value_range
        }
        return self
    
    def set_colors(self, *colors):
        self.color_range = list(colors)
        return self
    
    def set_connected_overlays(self, *overlays):
        self.connected_overlays = list(overlays)
        return self

    def to_dict(self):
        out = {
            "id": self.title.lower(),
            "title": self.title,
            "type": self.type,
            "source": self.source,
            "info": self.info,
        }
        if self.access_token:
            out["accessToken"] = self.access_token
        if hasattr(self, "connectedOverlays"):
            out["connectedOverlays"] = self.connected_overlays
        if hasattr(self, "legend"):
            out["legend"] = self.legend
        if hasattr(self, "colorRange"):
            out["colorRange"] = self.color_range
        if hasattr(self, "bounds"):
            out["bounds"] = self.bounds
        return out


base_map_layer = BaseLayer("Base", "vector-tile")\
    .set_source(
        "mapbox://styles/biotia/clmgqcx2103oq01p96ifm4kda",
         "pk.eyJ1IjoiYmlvdGlhIiwiYSI6ImNsZnY2cDlzMzA0N2gzcHFrbmlza2FxZGsifQ.zIR-YJ2grxJBPyBdyxhRew"
    )

light_base_map_layer = BaseLayer("Light", "vector-tile")\
    .set_source(
        "mapbox://styles/biotia/clmd9upef03co01qihm0fgaz6",
        "pk.eyJ1IjoiYmlvdGlhIiwiYSI6ImNsZnY2cDlzMzA0N2gzcHFrbmlza2FxZGsifQ.zIR-YJ2grxJBPyBdyxhRew"
    )

geography_base_layer = BaseLayer("Geography", "raster-tile")\
    .set_source(
        "https://api.mapbox.com/styles/v1/mapbox/satellite-v9/tiles/{z}/{x}/{y}?access_token=pk.eyJ1IjoidWxxdWlyb2xhIiwiYSI6ImNsa2pnZDBpNDBiajMzZm1xYTlsdDZ5dTUifQ.9XiWdIBNzXe09uiBzo3www"
    )

temperature_base_layer = BaseLayer("Temperature", "image")\
    .set_source("/maps/NASA_20221101_MOD_LSTD_M_3600x1800_3857_color.png")\
    .set_legend("Temperature (°C)", [-25, 45])\
    .set_colors("#56a3f0", "#ebdf76", "#f9693b")\
    .set_bounds(-90, -180, 90, 180)\
    .set_info("Daytime land surface temperatures detected by Terra satellite's MODIS device in clear-sky conditions through November 2022 and processed by NASA Earth Observatory.")\
    .set_connected_overlays("administration", "places")

humidity_base_layer = BaseLayer("Humidity", "image")\
    .set_source("/maps/NASA_20230201_EO_Aqua_MODIS_Water_vapor_3600x1800_3857_color.png")\
    .set_legend("Humidity (cm)", [0, 6])\
    .set_colors("#ababab6b", "#ffffff", "#9cd8e9", "#2e8bb3", "#018571")\
    .set_bounds(-90, -180, 90, 180)\
    .set_info("5 km² atmospheric column water vapor estimates from clear sky moisture profiles through February 2023 of Aqua satellite's moderate resolution imaging spectroradiometer (MODIS) processed by NASA Earth Observatory.")\
    .set_connected_overlays("administration", "places")

precipitation_base_layer = BaseLayer("Precipitation", "image")\
    .set_source("/maps/NASA_20230201_EO_Precipitation_3600x1800_3857_color.png")\
    .set_legend("Average Precipitation (mm)", [0, 3000])\
    .set_colors("#E6E7E8", "#005645")\
    .set_bounds(-90, -180, 90, 180)\
    .set_info("Integrated multi-satellite retrievals (IMERGI) of liquid and solid precipitation through February 2023 from the Global Precipitation Measurement (GPM) Mission's Core Observatory satellite constellation.")\
    .set_connected_overlays("administration", "places")



from geoseeq import GeoseeqNotFoundError
from .from_uuids import (
    org_from_uuid,
    project_from_uuid,
    sample_from_uuid,
    sample_result_folder_from_uuid,
    project_result_folder_from_uuid,
    result_folder_from_uuid,
    sample_result_file_from_uuid,
    project_result_file_from_uuid,
    result_file_from_uuid,
)
from .from_names import (
    org_from_name,
    project_from_name,
    sample_from_name,
    sample_result_folder_from_name,
    project_result_folder_from_name,
    result_folder_from_name,
    sample_result_file_from_name,
    project_result_file_from_name,
    result_file_from_name,
)
from .utils import is_grn_or_uuid, is_name
from geoseeq.knex import with_knex


def _generic_from_id(knex, id, from_uuid_func, from_name_func):
    """Return the object which the id points to."""
    if is_grn_or_uuid(id):
        return from_uuid_func(knex, id)
    if is_name(id):
        return from_name_func(knex, id)
    raise ValueError(f'"{id}" is not a GRN, UUID, or name')


@with_knex
def org_from_id(knex, id):
    """Return the organization object which the id points to."""
    return _generic_from_id(knex, id, org_from_uuid, org_from_name)


@with_knex
def project_from_id(knex, id):
    """Return the project object which the id points to."""
    return _generic_from_id(knex, id, project_from_uuid, project_from_name)


@with_knex
def sample_from_id(knex, id):
    """Return the sample object which the id points to."""
    return _generic_from_id(knex, id, sample_from_uuid, sample_from_name)


@with_knex
def sample_result_folder_from_id(knex, id):
    """Return the sample result folder object which the id points to."""
    return _generic_from_id(knex, id, sample_result_folder_from_uuid, sample_result_folder_from_name)


@with_knex
def project_result_folder_from_id(knex, id):
    """Return the project result folder object which the id points to."""
    return _generic_from_id(knex, id, project_result_folder_from_uuid, project_result_folder_from_name)


@with_knex
def result_folder_from_id(knex, id):
    """Return a result folder object which the id points to.
    
    Guess the result folder is a sample result folder. If not, try a project result folder.
    """
    return _generic_from_id(knex, id, result_folder_from_uuid, result_folder_from_name)


@with_knex
def sample_result_file_from_id(knex, id):
    """Return the sample result file object which the id points to."""
    return _generic_from_id(knex, id, sample_result_file_from_uuid, sample_result_file_from_name)


@with_knex
def project_result_file_from_id(knex, id):
    """Return the project result file object which the id points to."""
    return _generic_from_id(knex, id, project_result_file_from_uuid, project_result_file_from_name)


@with_knex
def result_file_from_id(knex, id):
    """Return a result file object which the id points to.
    
    Guess the result file is a sample result file. If not, try a project result file.
    """
    return _generic_from_id(knex, id, result_file_from_uuid, result_file_from_name)







import json
from glob import glob
from string import Template
from os.path import join

from .vc_stub import VCStub


class VCResultPattern:
    """A class to connect filepaths on the local system to GeoSeeq Results.
    
    Available Patterns:
     - {parent_obj_name}
     - {result_name}
     - {result_replicate} (replicate number of the result)
     - {field_name}
     - {field_checksum} (checksum of the data stored in field)
     - {field_updated_at} (timestamp when the field was last updated)
     - {ext} (file extension)
    """
    DEFAULT_PATTERN = '${parent_obj_name}/${result_name}/${result_name}.${result_replicate}.${field_name}.${ext}'

    def __init__(self, pattern_str):
        self.pattern = pattern_str

    def to_path(self, field):
        """Return the path that should be used to store this field.

        Do not write the field to path.
        """
        template = self.pattern
        if '${parent_obj_name}' in template:
            template = Template(template).safe_substitute(parent_obj_name=field.parent.parent.name)
        if '${result_name}' in template:
            template = Template(template).safe_substitute(result_name=field.parent.module_name)
        if '${result_replicate}' in template:
            template = Template(template).safe_substitute(result_replicate=field.parent.replicate)
        if '${field_name}' in template:
            template = Template(template).safe_substitute(field_name=field.name)
        if '${ext}' in template:
            template = Template(template).safe_substitute(ext=field.get_referenced_filename_ext())
        return template

    def from_path(self, path):
        """Return an AnalysisResult and Field generated from this path.
        
        Do not create the field on the remote or attempt to fetch matching data.
        """
    pass


class VCSample:
    """A class to represent a sample on GeoSeeq via version control.
    
    A version controlled sample can do these things:
     - create stub files locally for new results on the remote
     - create stub files locally for local files that match a particular pattern
    """

    def __init__(self, sample, result_dir, result_pattern=VCResultPattern.DEFAULT_PATTERN):
        self.sample = sample
        self.result_dir = result_dir
        self.result_pattern = VCResultPattern(result_pattern)

    def save_settings(self, geoseeq_dir):
        """Save the setting for this VCSample to disk."""
        blob = {
            "brn": self.sample.brn,  # BRN for the sample
            "results_dir": self.result_dir,  # directory where results are stored
        }
        if self.result_pattern:
            blob["result_pattern"] = self.result_pattern.pattern  # how geoseeq result names should be mapped to filepaths
        else:
            blob["result_pattern"] = VCResultPattern.DEFAULT_PATTERN
        filepath = join(geoseeq_dir, 'sample.json')
        with open(filepath, 'w') as f_out:
            f_out.write(json.dumps(blob, indent=4))

    @classmethod
    def parse_settings(cls, filepath):
        """Return a `VCSample based on the settings filepath."""
        blob = json.loads(open(filepath).read())
        sample = resolve_brn(blob['brn'])
        return cls(
            sample,
            blob['result_dir'],
            result_pattern=blob.get('result_pattern', VCResultPattern.DEFAULT_PATTERN)
        )

    def stubs_from_remote(self):
        for result in self.sample.get_analysis_results():
            for field in result.get_fields():
                path = self.result_pattern.to_path(field)
                stub = VCStub(field.brn, path, field.checksum())
                stub.set_parent_info(
                    self.sample.brn,
                    result.module_name,
                    result.replicate
                )
                yield stub, path




import time
import json
from os.path import basename, getsize
from pathlib import Path

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.utils import md5_checksum
from concurrent.futures import ThreadPoolExecutor, as_completed
from .utils import *


class FileChunker:

    def __init__(self, filepath, chunk_size):
        self.filepath = filepath
        self.chunk_size = chunk_size
        self.file_size = getsize(filepath)
        self.n_parts = int(self.file_size / self.chunk_size) + 1
        self.loaded_parts = []

    def load_all_chunks(self):
        if len(self.loaded_parts) != self.n_parts:
            with open(self.filepath, "rb") as f:
                f.seek(0)
                for i in range(self.n_parts):
                    chunk = f.read(self.chunk_size)
                    self.loaded_parts.append(chunk)
        return self  # convenience for chaining

    def get_chunk(self, num):
        self.load_all_chunks()
        return self.loaded_parts[num]
    
    def get_chunk_size(self, num):
        self.load_all_chunks()
        return len(self.loaded_parts[num])


class ResultFileUpload:
    """Abstract class that handles upload methods for result files."""

    def _create_multipart_upload(self, filepath, file_size, optional_fields):
        optional_fields = optional_fields if optional_fields else {}
        optional_fields.update(
            {
                "md5_checksum": md5_checksum(filepath),
                "file_size_bytes": file_size,
            }
        )
        data = {
            "filename": basename(filepath),
            "optional_fields": optional_fields,
            "result_type": "sample" if self.is_sample_result else "group",
        }
        response = self.knex.post(f"/ar_fields/{self.uuid}/create_upload", json=data)
        return response
    
    def _prep_multipart_upload(self, filepath, file_size, chunk_size, optional_fields):
        n_parts = int(file_size / chunk_size) + 1
        response = self._create_multipart_upload(filepath, file_size, optional_fields)
        upload_id = response["upload_id"]
        parts = list(range(1, n_parts + 1))
        data = {
            "parts": parts,
            "stance": "upload-multipart",
            "upload_id": upload_id,
            "result_type": "sample" if self.is_sample_result else "group",
        }
        response = self.knex.post(f"/ar_fields/{self.uuid}/create_upload_urls", json=data)
        urls = response
        return upload_id, urls
    
    def _upload_one_part(self, file_chunker, url, num, max_retries, session=None):
        file_chunk = file_chunker.get_chunk(num)
        attempts = 0
        while attempts < max_retries:
            try:
                if session:
                    http_response = session.put(url, data=file_chunk)
                else:
                    http_response = requests.put(url, data=file_chunk)
                http_response.raise_for_status()
                logger.debug(f"Upload for part {num + 1} succeeded.")
                break
            except requests.exceptions.HTTPError:
                logger.warn(
                    f"Upload for part {num + 1} failed. Attempt {attempts + 1} of {max_retries}."
                )
                attempts += 1
                if attempts == max_retries:
                    raise
                time.sleep(10**attempts)  # exponential backoff, (10 ** 2)s default max
        return {"ETag": http_response.headers["ETag"], "PartNumber": num + 1}
    
    def _finish_multipart_upload(self, upload_id, complete_parts):
        response = self.knex.post(
            f"/ar_fields/{self.uuid}/complete_upload",
            json={
                "parts": complete_parts,
                "upload_id": upload_id,
                "result_type": "sample" if self.is_sample_result else "group",
            },
            json_response=False,
        )
        response.raise_for_status()

    def _upload_parts(self, file_chunker, urls, max_retries, session, progress_tracker, threads):
        if threads == 1:
            logger.info(f"Uploading parts in series for {file_chunker.filepath}")
            complete_parts = []
            for num, url in enumerate(list(urls.values())):
                response_part = self._upload_one_part(file_chunker, url, num, max_retries, session)
                complete_parts.append(response_part)
                if progress_tracker: progress_tracker.update(file_chunker.get_chunk_size(num))
                logger.info(f'Uploaded part {num + 1} of {len(urls)} for "{file_chunker.filepath}"')
            return complete_parts
        
        with ThreadPoolExecutor(max_workers=threads) as executor:
            logger.info(f"Uploading parts in parallel for {file_chunker.filepath} with {threads} threads.")
            futures = []
            for num, url in enumerate(list(urls.values())):
                future = executor.submit(
                    self._upload_one_part, file_chunker, url, num, max_retries, session
                )
                futures.append(future)
            complete_parts = []
            for future in as_completed(futures):
                response_part = future.result()
                complete_parts.append(response_part)
                if progress_tracker: progress_tracker.update(file_chunker.get_chunk_size(response_part["PartNumber"] - 1))
                logger.info(
                    f'Uploaded part {response_part["PartNumber"]} of {len(urls)} for "{file_chunker.filepath}"'
                )
        complete_parts = sorted(complete_parts, key=lambda x: x["PartNumber"])
        return complete_parts

    def multipart_upload_file(
        self,
        filepath,
        file_size,
        optional_fields=None,
        chunk_size=FIVE_MB,
        max_retries=3,
        session=None,
        progress_tracker=None,
        threads=1,
    ):
        """Upload a file to S3 using the multipart upload process."""
        logger.info(f"Uploading {filepath} to S3 using multipart upload.")
        upload_id, urls = self._prep_multipart_upload(filepath, file_size, chunk_size, optional_fields)
        logger.info(f'Starting upload for "{filepath}"')
        complete_parts = []
        file_chunker = FileChunker(filepath, chunk_size).load_all_chunks()
        if progress_tracker: progress_tracker.set_num_chunks(file_chunker.file_size)
        complete_parts = self._upload_parts(file_chunker, urls, max_retries, session, progress_tracker, threads)
        self._finish_multipart_upload(upload_id, complete_parts)
        logger.info(f'Finished Upload for "{filepath}"')
        return self

    def upload_file(self, filepath, multipart_thresh=FIVE_MB, **kwargs):
        resolved_path = Path(filepath).resolve()
        file_size = getsize(resolved_path)
        return self.multipart_upload_file(filepath, file_size, **kwargs)
    
    def upload_json(self, data, **kwargs):
        """Upload a file with the given data as JSON."""
        with NamedTemporaryFile("w", suffix='.json') as f:
            json.dump(data, f)
            f.flush()
            return self.upload_file(f.name, **kwargs)



from typing import Literal
from .base_layer import (
    BaseLayer,
    base_map_layer,
    light_base_map_layer,
    geography_base_layer,
    temperature_base_layer,
    humidity_base_layer,
    precipitation_base_layer
)
from .overlay import (
    Overlay,
    admin_overlay,
    places_overlay,
)


class Map:
    """Specify an interactive map plot on GeoSeeq."""

    def __init__(self) -> None:
        self.base_layers = []
        self.overlays = []
        self.samples = "project"
        self.show_histo = True
        self.center = [0, 0]
        self.zoom = 2
        pass
    
    def set_center(self, lat, lon, zoom=2):
        """Set the map zoom by setting the center and zoom level."""
        self.center = [lat, lon]
        self.zoom = zoom
        return self
    
    def set_zoom(self, zoom):
        """Set the map zoom level."""
        self.zoom = zoom
        return self
    
    def add_custom_background_layer(self, base_layer: BaseLayer):
        """Add a custom background layer to the map."""
        self.base_layers.append(base_layer)
        return self
    
    def add_base_map(self):
        """Add a light blue base map layer to the map."""
        self.base_layers.append(base_map_layer)
        return self
    
    def add_light_base_map(self):
        """Add a white base map layer to the map."""
        self.base_layers.append(light_base_map_layer)
        return self
    
    def add_geography_base_map(self):
        """Add a base map layer with sattelite geography to the map."""
        self.base_layers.append(geography_base_layer)
        return self
    
    def add_temperature_base_map(self):
        """Add a base map layer with temperature to the map."""
        self.base_layers.append(temperature_base_layer)
        return self
    
    def add_humidity_base_map(self):
        """Add a base map layer with humidity to the map."""
        self.base_layers.append(humidity_base_layer)
        return self
    
    def add_precipitation_base_map(self):
        """Add a base map layer with precipitation to the map."""
        self.base_layers.append(precipitation_base_layer)
        return self
    
    def add_administrative_overlay(self):
        """Add an overlay with administrative boundaries to the map."""
        self.overlays.append(admin_overlay)
        return self
    
    def add_places_overlay(self):
        """Add an overlay with place names to the map."""
        self.overlays.append(places_overlay)
        return self
    
    def add_custom_overlay(self, overlay: Overlay):
        """Add a custom overlay to the map."""
        self.overlays.append(overlay)
        return self
    
    def add_project_samples(self):
        return self
    
    def add_all_samples(self):
        return self
    
    def set_histogram(self, visible: bool=True, color: str="#BBC8E3"):
        self.show_histo = visible
        self.histo_color = color
        return self
    
    def set_samples(self, samples: Literal["project", "all", "none"], color: str="#004ae4"):
        self.samples = samples
        self.samples_color = color
        return self
    
    def to_dict(self):
        out = {
            "type": "geoseeq-map",
            "version": 1,
            "settings": {
                "center": self.center,
                "flyToInit": False,
                "zoom": self.zoom,
            },
            "baseLayers": [bl.to_dict() for bl in self.base_layers],
            "overlays": [ol.to_dict() for ol in self.overlays],
            "histogram": {
                "display": self.show_histo
            },
            "samples": {
                "display": self.samples
            }
        }
        if hasattr(self, "histo_color"):
            out["histogram"]["color"] = self.histo_color
        if hasattr(self, "samples_color"):
            out["samples"]["color"] = self.samples_color
        return out



import click

from .upload import (
    cli_upload_file,
    cli_metadata,
)
from .upload_reads import cli_upload_reads_wizard
from .upload_advanced import cli_find_urls_for_reads

@click.group('upload')
def cli_upload():
    """Upload files to GeoSeeq."""
    pass

cli_upload.add_command(cli_upload_reads_wizard)
cli_upload.add_command(cli_upload_file)
cli_upload.add_command(cli_metadata)

@click.group('upload')
def cli_upload_advanced():
    """Advanced tools to upload files to GeoSeeq."""
    pass

cli_upload_advanced.add_command(cli_find_urls_for_reads)


from typing import Literal


class Overlay:

    def __init__(self, title: str, type: Literal["geojson", "vector-tile"]) -> None:
        self.title = title
        self.type = type
        self.source = None
        self.access_token = None
        self.info = ""
        self.style = {}
    
    def set_source(self, source: str, access_token: str=None):
        self.source = source
        if access_token:
            self.access_token = access_token
        return self
    
    def set_info(self, info: str):
        self.info = info
        return self
    
    def set_style(self, weight: float=1, fill_opacity: float=0, color: str="#737373"):
        self.style = {
            "weight": weight,
            "fillOpacity": fill_opacity,
            "color": color
        }
        return self
    
    def to_dict(self):
        out = {
            "id": self.title.lower(),
            "title": self.title,
            "type": self.type,
            "source": self.source,
            "info": self.info
        }
        if self.access_token:
            out["accessToken"] = self.access_token
        if self.style:
            out["style"] = self.style
        return out
    

admin_overlay = Overlay("Administration", "geojson")\
    .set_source("/maps/custom.geo.json")\
    .set_style(weight=1, fill_opacity=0, color="#737373")\
    .set_info("Static administrative boundaries. World Bank, Creative Commons Attribution 4.0 license, https://datacatalog.worldbank.org/search/dataset/0038272")

places_overlay = Overlay("Places", "vector-tile")\
    .set_source(
        "mapbox://styles/biotia/cllh2jweb01yq01mf5yl82s5e",
        "pk.eyJ1IjoiYmlvdGlhIiwiYSI6ImNsZnY2cDlzMzA0N2gzcHFrbmlza2FxZGsifQ.zIR-YJ2grxJBPyBdyxhRew"
    )\
    .set_info("Country, state, settlement labels")







class User:

    def __init__(self, knex, email):
        self.knex = knex
        self.email = email

    @property
    def username(self):
        return self.email

    # New endpoint has to be created on server and registration can be done with clerk backend API
    # But login cannot be done like that

    # def register(self):
    #     response = requests.post(
    #         f'{self.knex.endpoint_url}/auth/users/',
    #         headers=self.knex.headers,
    #         json={
    #             'email': self.username,
    #             'password': self.password,
    #         }
    #     )
    #     response.raise_for_status()
    #     self.login()
    #     return self



import logging
import click
import requests
from os.path import basename, getsize
from .upload_reads import (
    _make_in_process_logger,
    _get_regex,
    _group_files,
    flatten_list_of_fastqs,
)

from multiprocessing import Pool, current_process

from geoseeq.cli.constants import *
from geoseeq.cli.shared_params import (
    handle_project_id,
    private_option,
    module_option,
    project_id_arg,
    overwrite_option,
    yes_option,
    use_common_state,
)

from geoseeq.constants import FASTQ_MODULE_NAMES
from geoseeq.cli.progress_bar import PBarManager

logger = logging.getLogger('geoseeq_api')


def _keep_only_authentication_url_args(url):
    """Return a url with only the S3 authentication args"""
    root, args = url.split('?')
    args = args.split('&')
    args = [arg for arg in args if arg.startswith('AWSAccessKeyId=') or arg.startswith('Signature=')]
    return root + '?' + '&'.join(args)


def _get_url_for_one_file(args):
    """Return a tuple of the filepath and the url to upload it to"""
    result_file, filepath, overwrite, log_level = args
    _make_in_process_logger(log_level)
    if result_file.exists() and not overwrite:  
        return
    result_file = result_file.idem()
    file_size = getsize(filepath)
    _, urls = result_file._prep_multipart_upload(filepath, file_size, file_size + 1, {})
    url = _keep_only_authentication_url_args(urls['1'])
    return filepath, url


def _find_target_urls(groups, module_name, lib, filepaths, overwrite, cores, state):
    """Use GeoSeeq to get target urls for a set of files"""
    with requests.Session() as session:
        find_url_args = []
        for group in groups:
            sample = lib.sample(group['sample_name']).idem()
            read_folder = sample.result_folder(module_name).idem()

            for field_name, path in group['fields'].items():
                result_file = read_folder.read_file(field_name)
                filepath = filepaths[path]
                find_url_args.append((
                    result_file, filepath, overwrite, state.log_level
                ))

        with Pool(cores) as p:
            for (file_name, target_url) in p.imap_unordered(_get_url_for_one_file, find_url_args):
                yield file_name, target_url


@click.command('read-links')
@use_common_state
@click.option('--cores', default=1, help='Number of uploads to run in parallel')
@overwrite_option
@yes_option
@click.option('--regex', default=None, help='An optional regex to use to extract sample names from the file names')
@private_option
@module_option(FASTQ_MODULE_NAMES)
@project_id_arg
@click.argument('fastq_files', type=click.Path(exists=True), nargs=-1)
def cli_find_urls_for_reads(state, cores, overwrite, yes, regex, private, module_name, project_id, fastq_files):
    """Print a two column list with filenames and a target storage URL
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes, private)
    filepaths = {basename(line): line for line in flatten_list_of_fastqs(fastq_files)}
    click.echo(f'Found {len(filepaths)} files to upload.', err=True)
    regex = _get_regex(knex, filepaths, module_name, proj, regex)
    groups = _group_files(knex, filepaths, module_name, regex, yes)
    for file_name, target_url in _find_target_urls(groups, module_name, proj, filepaths, overwrite, cores, state):
        print(f'{file_name}\t{target_url}', file=state.outfile)



from .map import Map
from .base_layer import (
    BaseLayer,
    base_map_layer,
    light_base_map_layer,
    geography_base_layer,
    temperature_base_layer,
    humidity_base_layer,
    precipitation_base_layer,
)
from .overlay import (
    Overlay,
    admin_overlay,
    places_overlay,
)


from geoseeq import GeoseeqNotFoundError
from geoseeq.knex import with_knex


@with_knex
def org_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return an Organization object from a blob."""
    from geoseeq.organization import Organization  # import here to avoid circular import
    org = Organization(knex, blob["name"])
    org.load_blob(blob)
    org._already_fetched = already_fetched
    org._modified = modified
    return org


@with_knex
def project_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a Project object from a blob."""
    org = org_from_blob(
        knex, blob["organization_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.project import Project  # import here to avoid circular import
    grp = Project(knex, org, blob["name"], is_library=blob["is_library"])
    grp.load_blob(blob)
    grp._already_fetched = already_fetched
    grp._modified = modified
    return grp


sample_group_from_blob = project_from_blob  # Alias


@with_knex
def sample_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a Sample object from a blob."""
    lib = sample_group_from_blob(
        knex, blob["library_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.sample import Sample  # import here to avoid circular import
    sample = Sample(knex, lib, blob["name"], metadata=blob["metadata"])
    sample.load_blob(blob)
    sample._already_fetched = already_fetched
    sample._modified = modified
    return sample


@with_knex
def project_result_folder_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a ProjectResultFolder object from a blob."""
    group = project_from_blob(
        knex, blob["sample_group_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.result import ProjectResultFolder  # import here to avoid circular import
    ar = ProjectResultFolder(
        knex, group, blob["module_name"], replicate=blob["replicate"], metadata=blob["metadata"]
    )
    ar.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return ar


sample_group_ar_from_blob = project_result_folder_from_blob  # Alias


@with_knex
def sample_result_folder_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a SampleResultFolder object from a blob."""
    sample = sample_from_blob(
        knex, blob["sample_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.result import SampleResultFolder  # import here to avoid circular import
    ar = SampleResultFolder(
        knex, sample, blob["module_name"], replicate=blob["replicate"], metadata=blob["metadata"]
    )
    ar.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return ar


sample_ar_from_blob = sample_result_folder_from_blob  # Alias


@with_knex
def sample_result_file_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a SampleResultFile object from a blob."""
    ar = sample_result_folder_from_blob(
        knex, blob["analysis_result_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.result import SampleResultFile  # import here to avoid circular import
    arf = SampleResultFile(knex, ar, blob["name"], data=blob["stored_data"])
    arf.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return arf


sample_ar_field_from_blob = sample_result_file_from_blob  # Alias


@with_knex
def project_result_file_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a ProjectResultFile object from a blob."""
    ar = project_result_folder_from_blob(
        knex, blob["analysis_result_obj"], already_fetched=already_fetched, modified=modified
    )
    from geoseeq.result import ProjectResultFile  # import here to avoid circular import
    arf = ProjectResultFile(knex, ar, blob["name"], data=blob["stored_data"])
    arf.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return arf


sample_group_ar_field_from_blob = project_result_file_from_blob  # Alias


@with_knex
def pipeline_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a Pipeline object from a blob."""
    from geoseeq.pipeline import Pipeline  # import here to avoid circular import
    pipeline = Pipeline(knex, blob["name"])
    pipeline.load_blob(blob)
    pipeline._already_fetched = already_fetched
    pipeline._modified = modified
    return pipeline

app_from_blob = pipeline_from_blob  # Alias


@with_knex
def pipeline_run_from_blob(knex, blob, already_fetched=True, modified=False):
    """Return a Pipeline run object from a blob."""
    from geoseeq.pipeline import PipelineRun  # import here to avoid circular import
    pipeline_run = PipelineRun(
        knex,
        blob["sample_group"],
        blob["pipeline"],
        pipeline_version=blob["pipeline_version"],
    )
    pipeline_run.load_blob(blob)

    pipeline_run._already_fetched = already_fetched
    pipeline_run._modified = modified
    return pipeline_run



from .remote_object import RemoteObject
from typing import Literal


class Pipeline(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "description",
        "long_description",
        "input_elements",
    ]
    parent_field = None

    def __init__(self, knex, name):
        super().__init__(self)
        self.knex = knex
        self.name = name
        self.description = ""
        self.long_description = ""
        self._pipeline_options = []

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        url = f"pipelines/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f"pipelines/name/{self.name}")
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        data = {
            "name": self.name,
            "description": self.description,
            "long_description": self.long_description,
        }
        url = "pipelines?format=json"
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def __str__(self):
        return f"<Geoseeq::Pipeline {self.name} {self.uuid} />"

    def __repr__(self):
        return f"<Geoseeq::Pipeline {self.name} {self.uuid} />"

    def pre_hash(self):
        return "PIPELINE" + self.name

    def module(self, name, version, metadata={}):
        return PipelineModule(
            self.knex,
            self,
            name,
            version,
            metadata=metadata,
        )

    def get_modules(self):
        url = f"pipeline_modules?pipeline={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.module(result_blob["name"], result_blob["version"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result

    def options(self):
        """Return a list of PipelineOption objects for this pipeline."""
        if not self._pipeline_options:
            self._pipeline_options = [PipelineOption.from_blob(x) for x in self.input_elements]
        return self._pipeline_options
    
    def set_option(self, opt_name, opt_val):
        """Set the value of an option for this pipeline."""
        for opt in self.options():
            if opt.name == opt_name:
                opt.set_input(opt_val)
                return
        raise ValueError(f"Unknown option: {opt_name}")
    
    def get_input_parameters(self):
        """Return a dict of input parameters for this pipeline."""
        return {x.name: x.input_val for x in self.options()}


class PipelineOption:

    def __init__(self, name, type, default, description, default_value, options=[]):
        self.name = name
        self.type = type
        self.default = default
        self.description = description
        self.default_value = default_value
        self.options = options
        self._input_val = None

    def validate_input(self, input_val):
        """Return True in input_val is valid for this option."""
        if self.type == "checkbox":
            return input_val in [True, False]
        elif self.type == "select":
            return input_val in self.options
        elif self.type == "text":
            return isinstance(input_val, str)
        else:
            raise ValueError(f"Unknown option type: {self.type}")
    
    def set_input(self, input_val):
        """Set the input value for this option."""
        if not self.validate_input(input_val):
            raise ValueError(f"Invalid input value: {input_val}")
        self._input_val = input_val

    @property
    def input_val(self):
        """Return the input value for this option."""
        if self._input_val is None:
            return self.default_value
        return self._input_val
    
    @property
    def is_set(self):
        """Return True if this option has been set."""
        return self._input_val is not None
    
    def to_readable_description(self):
        """Return a human friendly description of this option."""
        out = f"{self.name} ({self.type})\n\t{self.description}\n\t"
        if self.type == "checkbox":
            out += f"(default: {self.default_value})"
        elif self.type == "select":
            out += f"options: {', '.join(self.options)}\n\t(default: {self.default_value})"
        elif self.type == "text":
            out += f"(default: {self.default_value})"
        else:
            raise ValueError(f"Unknown option type: {self.type}")
        return out
    
    @classmethod
    def from_blob(cls, blob):
        opts = [x['value'] for x in blob.get("options", {})]
        return cls(
            blob["name"],
            blob["type"],
            blob["default_value"],
            blob["description"],
            blob["default_value"],
            options=opts
        )




class PipelineModule(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "version",
        "metadata",
        "description",
        "long_description",
        "dependencies",
    ]
    parent_field = "pip"

    def __init__(self, knex, pipeline, name, version, metadata={}):
        super().__init__(self)
        self.knex = knex
        self.pip = pipeline
        self.name = name
        self.version = version
        self.metadata = metadata
        self.description = ""
        self.long_description = ""
        self.dependencies = []

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        url = f"pipeline_modules/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f"pipelines/{self.pip.uuid}/modules/{self.name}/{self.version}")
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        data = {
            "pipeline": self.pip.uuid,
            "name": self.name,
            "version": self.version,
            "metadata": self.metadata,
            "description": self.description,
            "long_description": self.long_description,
        }
        url = "pipeline_modules?format=json"
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def add_dependency(self, upstream):
        self.dependencies.append(upstream.uuid)

    def __str__(self):
        return f'<Geoseeq::PipelineModule "{self.name}" "{self.version}" {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::PipelineModule "{self.name}" "{self.version}" {self.uuid} />'

    def pre_hash(self):
        return "PIPELINE_MODULE" + self.name + self.version + self.pip.pre_hash()


class PipelineRun(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "sample_group",
        "sample",
        "pipeline",
        "pipeline_version",
        "user",
        "finished_at",
        "status",
        "phase",
        "error_message",
        "input_fields",
        "input_parameters",
    ]
    parent_field = None

    def __init__(
        self,
        knex,
        sample_group,
        pipeline,
        pipeline_version,
        user=None,
        sample=None,
        finished_at=None,
        status=None,
        phase=None,
        error_message=None,
        input_fields=None,
        input_parameters=None,
    ):
        super().__init__(self)
        self.knex = knex
        self.sample_group = sample_group
        self.pipeline = pipeline
        self.pipeline_version = pipeline_version
        self.sample = sample
        self.user = user
        self.finished_at = finished_at
        self.status = status
        self.phase = phase
        self.error_message = error_message
        self.input_fields = input_fields
        self.input_parameters = input_parameters

    def set_status(self, status: Literal["Running", "Finished", "Error", "Pending"]):
        """Set the status of the run and save it to the server. Returns self."""
        self.status = status
        self.save()
        return self

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        url = f"app_runs/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f"app_runs/{self.uuid}")
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        data = {
            "sample_group": self.sample_group,
            "pipeline": self.pipeline,
            "pipeline_version": self.pipeline_version,
            "sample": self.sample,
            "user": self.user,
            "finished_at": self.finished_at,
            "status": self.status,
            "phase": self.phase,
            "error_message": self.error_message,
            "input_fields": self.input_fields,
            "input_parameters": self.input_parameters,
        }
        url = f"app_runs"
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def pre_hash(self):
        return "PIPELINE_RUN" + self.sample_group



from .sample import Sample
from .knex import GeoseeqInternalError, GeoseeqTimeoutError
from time import sleep, time
import json
from .id_constructors import sample_from_id
import pandas as pd


class Search:

    def __init__(self, knex, search_terms: [str]=None, search_uuid=None) -> None:
        self.knex = knex
        self.search_terms = search_terms or []
        self.search_has_been_run = False
        self.search_uuid = search_uuid
        self.search_result = None
    
    def add_search_term(self, search_term: str) -> None:
        self.search_terms.append(search_term)
    
    def run_search(self):
        """Call the search API, save the results in this object, and return this object."""
        self._run_search()
        self.search_has_been_run = True
        return self
    
    def _init_search(self) -> None:
        """Post a call to the search API and save the search UUID."""
        if self.search_uuid: return  # search has already been initialized
        url = 'search'
        result = self.knex.post(url, json={'clauses': self.search_terms})
        self.search_uuid = result['uuid']

    def _poll_search(self) -> None:
        """Check if the search has finished. If it has return the response.
        
        If the search is pending return None.
        If the search has failed raise an exception.
        """
        url = f'search_result/{self.search_uuid}'
        result = self.knex.get(url)
        status = result['status']
        if status in ['pending', 'working']:
            return None
        elif status == 'error':
            raise GeoseeqInternalError('search failed')
        elif status == 'success':
            return result['result']
        else:  # unexpected status
            raise GeoseeqInternalError(f'Unexpected status: {status}')
        
    def _run_search(self, timeout_ms=100000, poll_interval_ms=100, poll_backoff=1.2) -> None:
        """Run the search and save the result."""
        self._init_search()
        poll_start_time = time()
        while True:
            result = self._poll_search()
            if result:
                self.search_result = result
                return
            if time() - poll_start_time > timeout_ms / 1000:
                raise GeoseeqTimeoutError(f'Search timed out after {timeout_ms} ms.')
            sleep(poll_interval_ms / 1000)
            poll_interval_ms *= poll_backoff

    def sample_table(self) -> pd.DataFrame:
        """Return a pandas dataframe with sample metadata."""
        if not self.search_has_been_run:
            self.run_search()
        rows = []
        for sample_blob in self.search_result['samples']:
            blob = sample_blob['metadata']
            blob['uuid'] = sample_blob['uuid']
            blob['sample_name'] = sample_blob['name']
            rows.append(blob)
        return pd.DataFrame(rows)

    def sample_uuids(self) -> [str]:
        """Return a list of sample UUIDs matching the search terms."""
        if not self.search_has_been_run:
            self.run_search()
        return [sample_blob['uuid'] for sample_blob in self.search_result['samples']]

    def samples(self):
        """Yield samples matching the search terms."""
        if not self.search_has_been_run:
            self.run_search()
        for sample_uuid in self.sample_uuids():
            yield sample_from_id(self.knex, sample_uuid)


import click
import json
from .shared_params import use_common_state


def eula_section_to_markdown(section):
    """Return a markdown string"""
    out = ""
    if section.get('title', None):
        out += "## {}\n\n".format(section['title'])
    for paragraph in section.get('paragraphs', []):
        if section.get('as_list', False):
            out += "* {}\n".format(paragraph)
        else:
            out += "{}\n\n".format(paragraph)
    return out


@click.group('eula')
def cli_eula():
    """Commands for working with the EULA."""
    pass

@cli_eula.command('show')
@use_common_state
def cli_get_eula(state):
    """Write the EULA to stdout as a markdown document."""
    knex = state.get_knex()
    eula_blob = knex.get('eula')
    out = "# GeoSeeq End User License Agreement\n\n"
    for section in eula_blob['sections']:
        out += eula_section_to_markdown(section)
    out += "Effective Date: {}\n\n".format(eula_blob['effective_date'])
    print(out, file=state.outfile)





"""Utils to get token from config"""
from os import makedirs, environ
from os.path import join, exists
import json
from geoseeq.knex import DEFAULT_ENDPOINT

CONFIG_FOLDER = environ.get("XDG_CONFIG_HOME", join(environ["HOME"], ".config"))
CONFIG_DIR = environ.get("GEOSEEQ_CONFIG_DIR", join(CONFIG_FOLDER, "geoseeq"))
PROFILES_PATH = join(CONFIG_DIR, "profiles.json")


def set_profile(token, endpoint=DEFAULT_ENDPOINT, profile="", overwrite=False):
    """Write a profile to a config file.
    
    Raises KeyError if profile already exists.
    """
    if not exists(PROFILES_PATH):
        makedirs(CONFIG_DIR)
        with open(PROFILES_PATH, "w") as f:
            json.dump({}, f)
    with open(PROFILES_PATH, "r") as f:
        profiles = json.load(f)
    profile = profile or "__default__"
    if profile in profiles and not overwrite:
        raise KeyError(f"Profile {profile} already exists.")
    profiles[profile] = {
        "token": token,
        "endpoint": endpoint,
    }
    with open(PROFILES_PATH, "w") as f:
        json.dump(profiles, f, indent=4)
    

def load_profile(profile=""):
    """Return an endpoit and a token"""
    profile = profile or "__default__"
    with open(PROFILES_PATH, "r") as f:
        profiles = json.load(f)
    if profile in profiles:
        return profiles[profile]["endpoint"], profiles[profile]["token"]
    raise KeyError(f"Profile {profile} not found.")



from .remote_object import RemoteObject


class App(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "description",
        "long_description",
        "organization_obj",
        "is_public",
        "input_elements",
    ]
    parent_field = None

    def __init__(self, knex, uuid):
        super().__init__(self)
        self.knex = knex
        self.uuid = uuid
        self.name = ""
        self.description = ""
        self.long_description = ""

    def get_post_data(self):
        data = {
            field: getattr(self, field)
            for field in self.remote_fields if hasattr(self, field)
        }
        return data

    def _save(self):
        data = self.get_post_data()
        url = f"pipelines/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f"pipelines/{self.uuid}")
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob, allow_overwrite=allow_overwrite)

    def _create(self):
        data = self.get_post_data()
        url = "pipelines?format=json"
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def __str__(self):
        return f"<Geoseeq::App {self.name} {self.uuid} />"

    def __repr__(self):
        return f"<Geoseeq::App {self.name} {self.uuid} />"

    def pre_hash(self):
        return "APP" + self.name

    # def module(self, name, version, metadata={}):
    #     return PipelineModule(
    #         self.knex,
    #         self,
    #         name,
    #         version,
    #         metadata=metadata,
    #     )

    # def get_modules(self):
    #     url = f"pipeline_modules?pipeline={self.uuid}"
    #     result = self.knex.get(url)
    #     for result_blob in result["results"]:
    #         result = self.module(result_blob["name"], result_blob["version"])
    #         result.load_blob(result_blob)
    #         # We just fetched from the server so we change the RemoteObject
    #         # meta properties to reflect that
    #         result._already_fetched = True
    #         result._modified = False
    #         yield result



class PlotSelector:

    def __init__(self, title, kind='dropdown', options=None):
        self.title = title
        assert kind in ['dropdown', 'radio']
        self.kind = kind
        self.options = options or []

    def __iter__(self):
        return iter(self.options)

    def add_option(self, option):
        if option not in self.options:
            self.options.append(option)

    def to_dict(self):
        assert self.options
        return {
            'title': self.title,
            'kind': self.kind,
            'options': self.options,
        }


class SelectablePlot:
    VERSION = 'v0'

    def __init__(self):
        self.selectors = []
        self.data = {}
    
    def add_selector(self, title, kind='dropdown', options=None):
        """Add a selector to the selectable plot. Return the selectable plot for chaining."""
        self.selectors.append(PlotSelector(title, kind=kind, options=options))
        return self

    def _add_keys(self, *keys):
        """Add keys to the data structure."""
        assert self.selectors
        assert len(keys) == len(self.selectors)
        subdata = self.data
        for i, key in enumerate(keys):
            if key not in subdata:
                subdata[key] = {}
                self.selectors[i].add_option(key)
            subdata = subdata[key]
    
    def add_plot(self, plot, *keys):
        """Add a plot to the selectable plot. Return the selectable plot for chaining."""
        self._add_keys(*keys)
        subdata = self.data
        for key in keys:
            if key == keys[-1]:
                subdata[key] = plot
            subdata = subdata[key]
        return self
    
    def _check_invariants(self):
        """Ensure we have plots for each key combo."""
        assert self.selectors
        
        def check_subdata(i, subdata, key_chain):
            if i == len(self.selectors):
                assert subdata, f"subdata does not exist or is empty. Key Chain: {key_chain}"
                return
            for key in self.selectors[i]:
                if key not in subdata:
                    assert False, f'key: "{key}" is missing in `data`'
                check_subdata(i + 1, subdata[key], key_chain + [key])
        
        check_subdata(0, self.data, [])

    def to_dict(self):
        """Return the selectable plot as a dictionary."""
        self._check_invariants()
        return {
            'version': self.VERSION,
            'selectors': [s.to_dict() for s in self.selectors],
            'data': self.data,
        }


import click

from .shared_params import (
    use_common_state,
    project_id_arg,
    pipeline_id_arg,
    handle_project_id,
    handle_pipeline_id,
)
from geoseeq.pipeline import PipelineOption

@click.group('app')
def cli_app():
    """Run pipelines on GeoSeeq."""
    pass


def prompt_for_option(pipeline_opt: PipelineOption):
    """Prompt the user for an option."""
    if pipeline_opt.is_set:
        return pipeline_opt.input_val
    if pipeline_opt.type == "checkbox":
        return click.confirm(pipeline_opt.description, default=pipeline_opt.default_value)
    elif pipeline_opt.type == "text":
        return click.prompt(pipeline_opt.description, default=pipeline_opt.default_value)
    elif pipeline_opt.type == "select":
        return click.prompt(
            pipeline_opt.description,
            default=pipeline_opt.default_value,
            type=click.Choice(pipeline_opt.options),
        )
    else:
        raise ValueError(f"Unknown option type: {pipeline_opt.type}")


@cli_app.command('run')
@use_common_state
@click.option('-po', '--pipeline-option', multiple=True, nargs=2, help='Set a pipeline option.')
@pipeline_id_arg
@project_id_arg
def run_pipeline(state, pipeline_option, pipeline_id, project_id):
    """Run a pipeline on all samples in a project.

    Some pipelines require options to be set. This command will prompt you to supply those options.
    Alternatively, you can set options with the `-po` flag. You can see required options using this
    command.

    `geoseeq view app "Pipeline Name"`

    ---

    Example Usage:

    \b
    # Run the "My Pipeline" pipeline on all samples in "My Org/My Project"
    $ geoseeq app run "My Pipeline" "My Org/My Project"

    \b
    # Run the "My Pipeline" pipeline on all samples in "My Org/My Project", setting the "my_option" option to "my_value"
    $ geoseeq app run "My Pipeline" "My Org/My Project" -po my_option my_value

    ---

    Command Arguments:

    \b
    [PIPELINE_ID] is the name or ID of the pipeline to run.
    [PROJECT_ID] is the name or ID of the project to run the pipeline on.

    ---

    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    knex = state.get_knex()
    project = handle_project_id(knex, project_id, create=False)
    pipeline = handle_pipeline_id(knex, pipeline_id)
    for opt_name, opt_val in pipeline_option:
        pipeline.set_option(opt_name, opt_val)
    for pipeline_opt in pipeline.options():
        pipeline_opt.set_input(prompt_for_option(pipeline_opt))
    project.run_app(pipeline)


@cli_app.command('status')
@use_common_state
@pipeline_id_arg
@project_id_arg
def get_pipeline_runs(state, pipeline_id, project_id):
    """See the status of all runs of a pipeline on a project.
    
    ---
    
    Example Usage:

    \b
    # See the status of all runs of "My Pipeline" on "My Org/My Project"
    $ geoseeq app status "My Pipeline" "My Org/My Project"

    ---

    Command Arguments:

    \b
    [PIPELINE_ID] is the name or ID of the pipeline to see the status of.
    [PROJECT_ID] is the name or ID of the project to see the status of.

    ---

    Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    knex = state.get_knex()
    project = handle_project_id(knex, project_id)
    pipeline_uuid = handle_pipeline_id(knex, pipeline_id).uuid
    response = knex.get(f'app_runs?sample_group_id={project.uuid}')
    for el in response['results']:
        if el['pipeline'] == pipeline_uuid:
            out = [
                el['pipeline_obj']['name'],
                el['uuid'],
                el['status'],
                el['created_at'],
            ]
            print('\t'.join(out), file=state.outfile)



import json
from os.path import join, dirname
from .checksum import Checksum
from os import environ, symlink, makedirs
from .vc_cache import VCCache


class VCStub:
    """A version control stub."""

    def __init__(self, brn, local_path, checksum, stub_path=None):
        self.brn = brn
        self._local_path = local_path
        self.checksum = checksum if isinstance(checksum, Checksum) else Checksum.from_blob(checksum)
        self.stub_path = stub_path
        self.parent_info = {}

    def download(self, knex):
        """Download the linked file to local storage."""
        _, field = self.field(knex)
        cache_dir = environ.get('GEOSEEQ_VC_CACHE_DIR', None)  # TODO diff caches for diff projects
        if cache_dir:
            cache = VCCache(cache_dir)
            cache_path = cache.get_cache_filepath(field)
            makedirs(dirname(cache_path), exist_ok=True)
            field.download_file(cache_path)
            symlink(cache_path, self.local_path)
        else:
            field.download_file(self.local_path)

    def verify(self):
        """Return True iff the local file matches the linked checksum."""
        return self.checksum.verify(self.local_path)

    def remote_verify(self, knex):
        """Return True iff the stub file matches the info on the remote."""
        pass

    def field(self, knex):
        """Return the AnalysisResultField that this stub points to on the remote."""
        return resolve_brn(knex, self.brn)
    
    @property
    def local_path(self):
        """Return the path on the local machine that this stub points to."""
        return join(dirname(self.stub_path), self._local_path)

    def set_parent_info(self, parent_obj_brn, result_module_name, result_replicate):
        self.parent_info = {
            "parent_obj_brn": parent_obj_brn,
            "result_module_name": result_module_name,
            "result_module_replicate": result_replicate,
        }

    def save_to_file(self, stub_path=None):
        assert stub_path or self.stub_path
        stub_path = stub_path if stub_path else self.stub_path
        makedirs(dirname(stub_path), exist_ok=True)
        with open(stub_path, 'w') as f_out:
            f_out.write(json.dumps(self.to_blob(), indent=4))

    def to_blob(self, parent_obj_brn=None, result_module_name=None, result_replicate=None):
        assert self.parent_info or self.brn
        blob = {
            "checksum": self.checksum.to_blob(),
            "local_path": self._local_path,
        }
        if self.brn:
            blob['brn'] = self.brn
        if self.parent_info:
            blob['parent_info'] = self.parent_info
        return blob

    @classmethod
    def from_blob(cls, blob, stub_path=None):
        assert 'brn' in blob or 'parent_info' in blob
        return cls(
            blob['brn'],
            blob['local_path'],
            Checksum.from_blob(blob['checksum']),
            stub_path=stub_path,
        )
    
    @classmethod
    def from_path(self, path):
        blob = json.loads(open(path).read())
        return VCStub.from_blob(blob, stub_path=path)



"""
Functions related to making highcharts plots on GeoSeeq.

GeoSeeq uses highcharts to make interactive plots.
"""

from .constants import COLORS


def series_is_categorical(col, max_unique=100, min_unique=2, min_not_null=3):
    """Return True if the series is categorical according to these criteria:
    1. The column is of type 'string'
    2. The column has lte 100 unique values
    3. The column has at least 2 unique values
    4. The column has at least 3 values that are not null
    """
    if col.dtype == 'object':
        if col.nunique() <= max_unique:
            if col.nunique() >= min_unique:
                if col.count() >= min_not_null:
                    return True
    return False


def series_is_numeric(col, min_not_null=3):
    """Return True if the series is numeric according to these criteria:
    1. The column can be cast to float
    2. The column has at least 3 values that are not null
    """
    try:
        col.astype(float)
        if col.count() >= min_not_null:
            return True
    except:
        pass
    return False


def _as_col_name(data, col_name):
    """Return the column name of data.
    
    If col_name is not a column name treat it as an index to the column list.
    """
    if col_name is None:
        return None
    if col_name in data.columns:
        return col_name
    else:
        return data.columns[col_name]


def scatter_plot_config(plot_title, data, x_col=0, y_col=1, color_col=None, x_title=None, y_title=None, color_title=None):
    """Return a highcharts config for a scatter plot of the data.
    
    If the color_col is categorical, the points will be colored by category.
    If the color_col is numeric, the points will be colored by a color scale.
    """
    x_col, y_col, color_col = _as_col_name(data, x_col), _as_col_name(data, y_col), _as_col_name(data, color_col)
    base_config = {
        "chart": {
            "type": 'scatter',
            "zoomType": 'xy',
        },
        "title": {
            "text": plot_title,
        },
        "xAxis": {
            "title": {
                "text": x_title or x_col,
            },
        },
        "yAxis": {
            "title": {
                "text": y_title or y_col,
            },
        },
        "tooltip": {
            "format": '<b>X:</b> {point.x} <b>Y:</b> {point.y}'
        },
        "plotOptions": {},
    }

    def to_pt_list(data):
        return list(data.T.to_dict(orient='list').values())
    
    if color_col:
        base_config["legend"] = {
            "title": {
                "text": color_title or color_col,
            },
        }
        base_config["tooltip"]["format"] = '<b>' + (color_title or color_col) + ': </b>{point.value}<br><b>X:</b> {point.x} <b>Y:</b> {point.y}'
        my_data = data[[x_col, y_col, color_col]].dropna()
        if series_is_categorical(data[color_col]):
            series = []
            for i, value in enumerate(my_data[color_col].unique()):
                value_data = my_data[my_data[color_col] == value]
                series.append({
                    'name': value,
                    'data': to_pt_list(value_data),
                    'color': COLORS[i % len(COLORS)],
                })
            base_config["series"] = series
        elif series_is_numeric(data[color_col]):
            minVal, maxVal = my_data[color_col].quantile(0.01) * 0.9, my_data[color_col].quantile(0.99) * 1.1
            base_config["colorAxis"] ={
                'minColor': COLORS[0], 'maxColor': COLORS[1], "min": minVal, "max": maxVal
            }
            base_config["series"] = [{
                "colorKey": "value",
                'name': color_title or color_col,
                'data': to_pt_list(my_data),
            }]
        else:
            raise ValueError(f'color_col "{color_col}" is not categorical or numeric')
    else:
        my_data = data[[x_col, y_col]].dropna()
        base_config["series"] = [{
            'name': 'Data',
            'data': to_pt_list(my_data),
            'color': COLORS[0],
        }]
    return base_config




import urllib.request
import logging
import requests
from os.path import basename, getsize, join, isfile
from pathlib import Path
from tempfile import NamedTemporaryFile

from geoseeq.utils import download_ftp
from geoseeq.constants import FIVE_MB

logger = logging.getLogger("geoseeq_api")  # Same name as calling module


def _download_head(url, filename, head=None, progress_tracker=None):
    headers = None
    if head and head > 0:
        headers = {"Range": f"bytes=0-{head}"}
    response = requests.get(url, stream=True, headers=headers)
    response.raise_for_status()
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    if progress_tracker: progress_tracker.set_num_chunks(total_size_in_bytes)
    block_size = FIVE_MB
    with open(filename, 'wb') as file:
        for data in response.iter_content(block_size):
            if progress_tracker: progress_tracker.update(len(data))
            file.write(data)
    return filename


def _download_generic(url, filename, head=None):
    urllib.request.urlretrieve(url, filename)
    return filename


def guess_download_kind(url):
    if 'azure' in url:
        return 'azure'
    elif 's3' in url:
        return 's3'
    elif 'ftp' in url:
        return 'ftp'
    else:
        return 'generic'


def download_url(url, kind='guess', filename=None, head=None, progress_tracker=None):
    """Return a local filepath to the downloaded file. Download the file."""
    if kind == 'guess':
        kind = guess_download_kind(url)
        logger.info(f"Guessed download kind: {kind} for {url}")
    logger.info(f"Downloading {kind} file to {filename}")
    if kind == 'generic':
        return _download_generic(url, filename, head=head)
    elif kind == 's3':
        return _download_head(url, filename, head=head, progress_tracker=progress_tracker)
    elif kind == 'azure':
        return _download_head(url, filename, head=head)
    elif kind == 'ftp':
        return download_ftp(url, filename, head=head)
    else:
        raise ValueError(f"Unknown download kind: {kind}")



class ResultFileDownload:
    """Abstract class that handles download methods for result files."""

    def get_download_url(self):
        """Return a URL that can be used to download the file for this result."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra", "ftp", "azure"]:
            raise ValueError(f'Unknown URL type: "{blob_type}"')
        key = 'url' if 'url' in self.stored_data else 'uri'
        if blob_type in ["s3", "azure"]:
            try:
                url = self.stored_data["presigned_url"]
            except KeyError:
                url = self.stored_data[key]
            if url.startswith("s3://"):
                url = self.stored_data["endpoint_url"] + "/" + url[5:]
            return url
        else:
            return self.stored_data[key]

    def download(self, filename=None, flag_suffix='.gs_downloaded', cache=True, head=None, progress_tracker=None):
        """Return a local filepath to the file in this result. Download the file if necessary.
        
        When the file is downloaded, it is cached in the result object. Subsequent calls to download
        on this object will return the cached file unless cache=False is specified.

        A flag file is created when the file download is complete. Subsequent calls to download
        will return the cached file if the flag file exists unless cache=False is specified.
        """
        if not filename and not not self._cached_filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        elif not filename and self._cached_filename:
            filename = self._cached_filename

        blob_type = self.stored_data.get("__type__", "").lower()
        if cache and self._cached_filename:
            return self._cached_filename
        flag_filename = filename + flag_suffix
        if cache and flag_suffix:
            # check if file and flag file exist, if so, return filename
            if isfile(filename) and isfile(flag_filename):
                return filename

        url = self.get_download_url()
        filepath = download_url(
            url, blob_type, filename,
            head=head, progress_tracker=progress_tracker
        )
        if cache and flag_suffix:
            # create flag file
            open(flag_filename, 'a').close()
        if cache:
            self._cached_filename = filepath
        return filepath



from os import environ
from os.path import join

FIVE_MB = 5 * (1024 ** 2)
FASTQ_MODULE_NAMES = ['short_read::paired_end', 'short_read::single_end', 'long_read::nanopore']
DEFAULT_ENDPOINT = "https://backend.geoseeq.com"

CONFIG_FOLDER = environ.get("XDG_CONFIG_HOME", join(environ["HOME"], ".config"))
CONFIG_DIR = environ.get("GEOSEEQ_CONFIG_DIR", join(CONFIG_FOLDER, "geoseeq"))
PROFILES_PATH = join(CONFIG_DIR, "profiles.json")


import click
import json
from .shared_params import (
    use_common_state,
    project_id_arg,
    sample_ids_arg,
    yes_option,
    private_option,
    org_id_arg,
    handle_project_id,
    handle_multiple_sample_ids,
    handle_org_id,
)
from geoseeq.id_constructors import resolve_id, pipeline_from_uuid


@click.group('detail')
def cli_detail():
    """Detail objects on GeoSeeq."""
    pass


def print_folder(folder, recurse=False):
    click.echo('Folder:')
    click.echo(folder)
    click.echo('Created at: {}'.format(folder.created_at))
    click.echo('Updated at: {}'.format(folder.updated_at))
    if recurse:
        for rfile in folder.get_result_files():
            click.echo(rfile)
            click.echo('Created at: {}'.format(rfile.created_at))
            click.echo('Updated at: {}'.format(rfile.updated_at))
            click.echo(json.dumps(rfile.stored_data, indent=2))
            click.echo('--')


def print_sample(sample, recurse=False):
    click.echo('Sample:')
    click.echo(sample)
    click.echo('Created at: {}'.format(sample.created_at))
    click.echo('Updated at: {}'.format(sample.updated_at))
    if recurse:
        for rfile in sample.get_result_folders():
            print_folder(rfile, recurse=recurse)


@cli_detail.command('folder')
@use_common_state
@click.argument('grn')
def detail_folder(state, grn):
    """Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    kind, rfolder = resolve_id(state.get_knex(), grn)
    assert kind == 'folder'
    click.echo('Folder:')
    click.echo(rfolder)
    click.echo('Created at: {}'.format(rfolder.created_at))
    click.echo('Updated at: {}'.format(rfolder.updated_at))
    click.echo('Files:')
    for rfile in rfolder.get_result_files():
        click.echo(rfile)
        click.echo('Created at: {}'.format(rfile.created_at))
        click.echo('Updated at: {}'.format(rfile.updated_at))
        click.echo(json.dumps(rfile.stored_data, indent=2))
        click.echo('--')


@cli_detail.command('project')
@use_common_state
@click.option('--recurse/--no-recurse', default=False)
@click.argument('grn')
def detail_project(state, recurse, grn):
    """Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    kind, project = resolve_id(state.get_knex(), grn)
    assert kind == 'project'
    click.echo('Project:')
    click.echo(project)
    click.echo('Privacy: {}'.format(project.privacy_level))
    click.echo('Created at: {}'.format(project.created_at))
    click.echo('Folders:')
    for folder in project.get_result_folders():
        print_folder(folder, recurse=recurse)
        click.echo('--')
    click.echo('Samples:')
    for sample in project.get_samples():
        print_sample(sample, recurse=recurse)
        click.echo('--')


@cli_detail.command('pipeline')
@use_common_state
@click.argument('grn')
def detail_pipeline(state, grn):
    """Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    pipeline_uuid = grn.split(':')[-1]
    pipeline = pipeline_from_uuid(state.get_knex(), pipeline_uuid)
    click.echo('Pipeline:')
    click.echo(pipeline)
    click.echo('Arguments:')
    for pipeline_option in pipeline.options():
        click.echo(f'Name: {pipeline_option.name}')
        click.echo(f'Type: {pipeline_option.type}')
        click.echo(f'Options: {pipeline_option.options}')
        click.echo(f'Default: {pipeline_option.default_value}')
        click.echo(f'Description: {pipeline_option.description}')
        click.echo('--')


@cli_detail.command('pipeline-run')
@use_common_state
@click.argument('uuid')
def detail_pipeline_run(state, uuid):
    """Use of this tool implies acceptance of the GeoSeeq End User License Agreement.
    Run `geoseeq eula show` to view the EULA.
    """
    pipeline_run = state.get_knex().get(f'app_runs/{uuid}')
    click.echo('Pipeline Run:')
    for key, val in pipeline_run.items():
        click.echo(f'{key}: {val}')
